#ifndef __NT2__CORE__DEFINITION__CONTAINER__DIMEN__ENTITY_HPP__INCLUDED
#define __NT2__CORE__DEFINITION__CONTAINER__DIMEN__ENTITY_HPP__INCLUDED

////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2003-2008 LASMEA UMR 6602 du CNRS.
// Copyright (c) 2007-2008 IEF    UMR 8622 du CNRS.
// All rights reserved.
//
// License information are available in the LICENSE file.
// Additionnal informations are available in the INFOS file.
////////////////////////////////////////////////////////////////////////////////

namespace nt2
{
  //////////////////////////////////////////////////////////////////////////////
  /// @brief size container
  /// dimen is a container that contains dimensiosn informations. Its main usage
  /// is to use NT2 ET functions to compute complex dimensions set.
  //////////////////////////////////////////////////////////////////////////////
  template<size_t N,class S>
  class dimen : public make_dimen<N,S>::type
  {
    public:

    ////////////////////////////////////////////////////////////////////////////
    // models ImplementPolicy concept
    ////////////////////////////////////////////////////////////////////////////
//    typedef settings settings_type(_2D::ofSize<1,N>,matlab_index);

    ////////////////////////////////////////////////////////////////////////////
    // Forward types from parent
    ////////////////////////////////////////////////////////////////////////////
    typedef typename make_dimen<N,S>::type       expr_type;
    typedef typename make_dimen<N,S>::base       base_type;

    ////////////////////////////////////////////////////////////////////////////
    // models Container concept
    ////////////////////////////////////////////////////////////////////////////
    typedef numel_t                              value_type;
    typedef typename base_type::reference_type   reference_type;
    typedef typename expr_type::size_type        size_type;

    ////////////////////////////////////////////////////////////////////////////
    // Canonic form construct/destruct/copy/assign
    ////////////////////////////////////////////////////////////////////////////
    dimen() {}
//    dimen( const type& src )
//    {
//      std::cout << "copy" << std::endl;
//      mSize = src.mSize;
//      mData = src.mData;
//    }
//
//    type& operator=( const type& src )
//    {
//      std::cout << "aff. copy" << std::endl;
//      if(this != &src)
//      {
//        mSize = src.mSize;
//        mData = src.mData;
//      }
//      return *this;
//    }

    ////////////////////////////////////////////////////////////////////////////
    /// @brief explicit constructors
    /// Builds a dimen from a set of 1 ... NT2_MAX_DIMENSIONS values
    /// A MPL static assert is thrown at compile time if the number of arguments
    /// is greater than the dimen instance dimension.NT2_TMP_MACRO
    ////////////////////////////////////////////////////////////////////////////
    #define NT2_TMP_MACRO(z,n,t) bp::value(*this)[n] = BOOST_PP_CAT(i,n);
    #define NT2_TMP_MACRO2(z,n,t)                                                       \
    explicit dimen( BOOST_PP_ENUM_PARAMS(n,const numel_t& i) )  \
    {                                                           \
      for(size_t i=n;i<N;++i) bp::value(*this)[i] = 1;          \
      BOOST_PP_REPEAT(n,NT2_TMP_MACRO,~)                        \
    }                                                           \
    /**/
    BOOST_PP_REPEAT_FROM_TO(1,NT2_PP_DIM_BOUND,NT2_TMP_MACRO2,~)
    #undef NT2_TMP_MACRO
    #undef NT2_TMP_MACRO2

//    ////////////////////////////////////////////////////////////////////////////
//    /// @brief Copy constructor from different dimen size
//    /// dimen of different size are projected onto a common subspace
//    /// before being copied. Basically, this means that extraneous dimen values
//    /// are folded and missing values are set to 1.
//    /// @param src dimen to copy
//    ////////////////////////////////////////////////////////////////////////////
//    template<size_t P,class X>
//    inline dimen( const dimen<P,X>&  src )
//    {
//      project(src);
//    }
//
//    template<size_t P,class X>
//    inline dimen& operator=( const dimen<P,X>&  src )
//    {
//      project(src);
//      return *this;
//    }
//
   ////////////////////////////////////////////////////////////////////////////
   /// @brief Oriented constructor.
   /// dimen is 1 except maybe in the along direction
   /// @param arg value/dimension pair to be used as non-trivial element
   /// @param value   value to be set in dimen along
   ////////////////////////////////////////////////////////////////////////////
   dimen(const numel_t & value, const along& arg)
   {
     for(size_t i=0;i<N;++i) bp::value(*this)[i] = 1;
     bp::value(*this)[arg.dimension()] = value;
   }

//    ////////////////////////////////////////////////////////////////////////////
//    /// @brief copy and modify constructor
//    ///
//    /// Build a dimen by copying the source dimen and modifying the along
//    /// value.
//    ///
//    /// @param src dimen original dimen
//    /// @param arg dimen whose value is to be set to val
//    /// @param value   value to be set in dimen along
//    ////////////////////////////////////////////////////////////////////////////
//    dimen(const type& src, const numel_t& value, const along& arg) /*: expr_type(mSize,mData)*/
//    {
//      mData = src.mData;
//      mData[arg.dimension()] = value;
//    }
//
//    ////////////////////////////////////////////////////////////////////////////
//    // Construct/copy from dimensions
//    ////////////////////////////////////////////////////////////////////////////
//    template<size_t M, class V>
//    dimen( dimensions<M,V> const& src ) /*: expr_type(mSize,mData)*/
//    {
//      project(src);
//    }
//
//    template<size_t M, class V>
//    inline dimen& operator=( dimensions<M,V> const&  src )
//    {
//      project(src);
//      return *this;
//    }
//
    ////////////////////////////////////////////////////////////////////////////
    // models Container concept
    ////////////////////////////////////////////////////////////////////////////
    value_type        operator()(const index_t& i)  const { return bp::value(*this)[i-1]; }
    reference_type    operator()(const index_t& i)        { return bp::value(*this)[i-1]; }
    const size_type&  size()                        const { return mSize; }

//    protected:
//    ////////////////////////////////////////////////////////////////////////////
//    // Copy dimen compatible data by colalpsing extraneous values or adding 1 in
//    // missing dimensions elements.
//    ////////////////////////////////////////////////////////////////////////////
//    template<size_t P,class X> inline void project( const dimen<P,X>& src )
//    {
//      for(size_t i=1              ;i<=(P>N ? N : P);++i)  mData[i-1]  = src[i];
//      for(size_t i=(P>N ? N+1 : 1);i<=(P>N ? P : 0);++i)  mData[N-1] *= src[i];
//      for(size_t i=(P<N ? P : 0)  ;i< (P<N ? N : 0);++i)  mData[i]    = 1;
//    }
//
//    template<size_t P,class X> inline void project( const dimensions<P,X>& src )
//    {
//      for(size_t i=0            ;i<(P>N ? N : P);++i) mData[i]    = src[i];
//      for(size_t i=(P>N ? N : 0);i<(P>N ? P : 0);++i) mData[N-1] *= src[i];
//      for(size_t i=(P<N ? P : 0);i<(P<N ? N : 0);++i) mData[i]    = 1;
//    }

    private:
    size_type mSize;
  };
}
#endif
