#ifndef __NT2__CORE__DEFINITION__CONTAINER__DIMEN__GRAMMAR_HPP__INCLUDED
#define __NT2__CORE__DEFINITION__CONTAINER__DIMEN__GRAMMAR_HPP__INCLUDED

////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2003-2008 LASMEA UMR 6602 du CNRS.
// Copyright (c) 2007-2008 IEF    UMR 8622 du CNRS.
// All rights reserved.
//
// License information are available in the LICENSE file.
// Additionnal informations are available in the INFOS file.
////////////////////////////////////////////////////////////////////////////////

#include <boost/proto/proto.hpp>
#include <boost/proto/transform.hpp>

namespace nt2
{
  //////////////////////////////////////////////////////////////////////////////
  // Dimen expression are built from :
  //  - dynamic dimen values
  //  - static dimen values
  //  - integral constants
  //  - arithmetic operations on dimen
  //////////////////////////////////////////////////////////////////////////////
  struct dimen_grammar : bp::or_<
    //////////////////////////////////////////////////////////////////////////
    // Terminals are dimensional values or integral values
    //////////////////////////////////////////////////////////////////////////
    bp::or_< bp::terminal< dimensions<bp::_,bp::_> >
           , integer_grammar
           >
    //////////////////////////////////////////////////////////////////////////
    // dimen can be added or scaled by an integer
    //////////////////////////////////////////////////////////////////////////    
   ,bp::or_< bp::plus<dimen_grammar,dimen_grammar>
	    ,bp::multiplies<dimen_grammar,integer_grammar>
            ,bp::multiplies<integer_grammar,dimen_grammar>
	    ,bp::shift_left<dimen_grammar,integer_grammar>
	    ,bp::shift_right<dimen_grammar,integer_grammar>
           >
    > {};

  template<class D> struct dimension_of
  {
    typedef bm::size_t<0> type;
  };

  template<class N, class S> struct dimension_of< dimensions<N,S> >
  {
    typedef N type;
  };

  //////////////////////////////////////////////////////////////////////////
  // Semantic actions for evaluating static size of dimen expression
  //////////////////////////////////////////////////////////////////////////
  struct dimen_size : bp::or_<
    //////////////////////////////////////////////////////////////////////////
      // dimensionnal values terminal contains its own size value
      //////////////////////////////////////////////////////////////////////////
    bp::when< bp::terminal< dimensions<bp::_,bp::_> >
	      , dimension_of< bp::_value>()
	        >
    //////////////////////////////////////////////////////////////////////////
    // integral values has a size of 1
    //////////////////////////////////////////////////////////////////////////
      , bp::when<integer_grammar, bm::size_t<1>()>
    //////////////////////////////////////////////////////////////////////////
    // size(d1+d2) = max(size(d1),size(d2))
    //////////////////////////////////////////////////////////////////////////    
      , bp::when< bp::plus<dimen_size,dimen_size>
		  , bm::max<dimen_size(bp::_left),dimen_size(bp::_right)>() 
      >
		      //////////////////////////////////////////////////////////////////////////
		    // size(d1*i) = size(d1)
		      //////////////////////////////////////////////////////////////////////////    
    , bp::when< bp::multiplies<dimen_size,integer_grammar>
	      , dimen_size(bp::_left)
              >
    //////////////////////////////////////////////////////////////////////////
    // size(i*d2) = size(d2)
    //////////////////////////////////////////////////////////////////////////    
    , bp::when< bp::multiplies<integer_grammar,dimen_size>
	      , dimen_size(bp::_right)
              >
    //////////////////////////////////////////////////////////////////////////
    // size(d1<<i) = size(d1)
    //////////////////////////////////////////////////////////////////////////    
    , bp::when< bp::shift_left<dimen_size,integer_grammar>
	      , dimen_size(bp::_left)
              >
    //////////////////////////////////////////////////////////////////////////
    // size(d1>>i) = size(d1)
    //////////////////////////////////////////////////////////////////////////    
    , bp::when< bp::shift_right<dimen_size,integer_grammar>
	      , dimen_size(bp::_left)
              >
    //////////////////////////////////////////////////////////////////////////
    // Other returns 0 but why ??? to fix it's fugly v_v 
    //////////////////////////////////////////////////////////////////////////    
    , bp::when<bp::_, bm::size_t<0>()>
    > {};

  template<typename X> struct size_
  {
    typedef typename boost::result_of<dimen_size(X)>::type                  base;
    typedef typename bm::apply1<OfSize<1,base::value>,bm::size_t<2> >::type type;
  };
}

#endif
