#ifndef __NT2__CORE__DEFINITION__CONTAINER__DIMEN__INNER_HPP__INCLUDED
#define __NT2__CORE__DEFINITION__CONTAINER__DIMEN__INNER_HPP__INCLUDED

////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2003-2008 LASMEA UMR 6602 du CNRS.
// Copyright (c) 2007-2008 IEF    UMR 8622 du CNRS.
// All rights reserved.
//
// License information are available in the LICENSE file.
// Additionnal informations are available in the INFOS file.
////////////////////////////////////////////////////////////////////////////////
#include <boost/mpl/empty_base.hpp>

namespace nt2
{
  //////////////////////////////////////////////////////////////////////////////
  /// The dynamic tag is used to flag run-time based container components
  //////////////////////////////////////////////////////////////////////////////
  struct dynamic_tag {};

  //////////////////////////////////////////////////////////////////////////////
  /// The static tag is used to flag compile-time based container components
  //////////////////////////////////////////////////////////////////////////////
  struct static_tag {};

  //////////////////////////////////////////////////////////////////////////////
  /// Internal dimensions container
  //////////////////////////////////////////////////////////////////////////////
  template<class N, class Values> struct dimensions
  {
    typedef typename Values::type values;

    ////////////////////////////////////////////////////////////////////////////
    /// status select if the current dimensions container is static or dynamic
    ////////////////////////////////////////////////////////////////////////////
    typedef is_same<values,void>                    status;

    ////////////////////////////////////////////////////////////////////////////
    /// Computes tag based on status
    ////////////////////////////////////////////////////////////////////////////
    typedef typename bm::if_<status,dynamic_tag,static_tag>::type    tag;

    ////////////////////////////////////////////////////////////////////////////
    /// Compute actual dimenions value storage based on status
    ////////////////////////////////////////////////////////////////////////////
    typedef typename bm::if_<status
                            ,array<size_t,N::value>
                            ,bm::empty_base
                            >::type                 data_type;

    ////////////////////////////////////////////////////////////////////////////
    /// Compute actual return type for operator[]
    ////////////////////////////////////////////////////////////////////////////
    typedef typename bm::if_<status,numel_t&,numel_t>::type  reference_type;
    typedef numel_t                                          value_type;
    ////////////////////////////////////////////////////////////////////////////
    /// How many dimensions are they ?
    ////////////////////////////////////////////////////////////////////////////
    typedef bm::size_t<N::value>           size_value;

    ////////////////////////////////////////////////////////////////////////////
    /// construct/copy/destruct
    ////////////////////////////////////////////////////////////////////////////
    dimensions() { reset(tag()); }
   ~dimensions() {}

    ////////////////////////////////////////////////////////////////////////////
    /// Direct element access
    /// @param i   : 0 based index
    ////////////////////////////////////////////////////////////////////////////
    reference_type  operator[](int i)       { return get(i,tag());  }
    value_type      operator[](int i) const { return cget(i,tag()); }
   
    protected:
    void reset(const static_tag& )  {}

    void reset(const dynamic_tag& )
    {
      mData[0] = 0;
      for(size_t i=1;i<N::value;++i) mData[i] = 1;
    }

    reference_type  get(int i, const dynamic_tag& )        
    { 
      static numel_t d = 1; 
      return (i>=0 && i<N::value) ? mData[i] : d;
    }
    
    value_type      cget(int i, const dynamic_tag& ) const
    { 
      return (i>=0 && i<N::value) ? mData[i] : 1;
    }

    reference_type  get(int i , const static_tag& s) { return cget(i,s); }
    value_type      cget(int i, const static_tag&  ) const
    {
      switch(i)
      {
        #define NT2_TMP_MACRO(z,n,text)                                \
	case n : return n<N::value                                     \
                      ? bm::at_c<values,(n<N::value?n:0)>::type::value \
                      : 1;                                             \
	/**/
        BOOST_PP_REPEAT(NT2_MAX_DIMENSIONS,NT2_TMP_MACRO,~)
        #undef NT2_TMP_MACRO
        default : return 1;
      }
    }

    private:
    data_type mData;
  };

  //////////////////////////////////////////////////////////////////////////////
  // Display dimensions on output streams
  //////////////////////////////////////////////////////////////////////////////
  template<class N, class Values> static inline
  std::ostream& operator<<( std::ostream& os, dimensions<N,Values> const& d )
  {
    os << "[";
    for(size_t i=0;i<N::value-1;++i) os << d[i] << ",";
    os << d[N::value-1] << "]";
    return os;
  }

  //////////////////////////////////////////////////////////////////////////////
  /// @brief dynamic dimension constructor meta-function
  //////////////////////////////////////////////////////////////////////////////
  struct default_dimensions
  {
    template<class N> struct apply
    {
      static const size_t dim = N::value;
      typedef dimensions<N> type;
    };
  };

  //////////////////////////////////////////////////////////////////////////////
  /// @brief static dimension constructor meta-function
  //////////////////////////////////////////////////////////////////////////////
  template<NT2_PP_ENUM_DIM_ARG> struct OfSize
  {
    typedef bm::vector_c<size_t,NT2_PP_ENUM_DIM>        type;

    template<size_t N,size_t NB=NT2_MAX_DIMENSIONS> struct make
    {
      typedef OfSize<NT2_PP_ENUM_DIM> type;
    };

    #define NT2_TMP_MACRO(z,n,text) BOOST_PP_CAT(D,n) *
    #define NT2_TMP_MACRO2(z,n,text)                                                \
    template<size_t NB> struct make<n,NB>                                           \
    {                                                                               \
      typedef OfSize< BOOST_PP_ENUM_PARAMS(BOOST_PP_DEC(n),D),                      \
                            BOOST_PP_REPEAT_FROM_TO(BOOST_PP_DEC(n),                \
                                                    NT2_PP_MAX_DIM,NT2_TMP_MACRO,~) \
                            BOOST_PP_CAT(D,NT2_PP_MAX_DIM)                          \
                          >                                                         \
                   type;                                                            \
    };

    template<size_t NB> struct make<1,NB>
    {
      typedef OfSize< BOOST_PP_REPEAT(NT2_PP_MAX_DIM,NT2_TMP_MACRO,~)
                              BOOST_PP_CAT(D,NT2_PP_MAX_DIM) >     type;
    };

    BOOST_PP_REPEAT_FROM_TO(2, BOOST_PP_INC(NT2_MAX_DIMENSIONS),NT2_TMP_MACRO2,~)
    #undef NT2_TMP_MACRO
    #undef NT2_TMP_MACRO2

    template<class N> struct apply
    {
      static const size_t dim = N::value;
      typedef dimensions<N,typename make<dim>::type> type;
    };
  };

  //////////////////////////////////////////////////////////////////////////////
  // Type evaluation for dimen parent
  //////////////////////////////////////////////////////////////////////////////
  template<size_t N,class S> struct make_dimen
  {
    typedef typename bm::apply1<S,bm::size_t<N> >::type   base;
    typedef dimen_expr<typename bp::terminal<base>::type> type;
  };
}

#endif
